from pathlib import Path

from chardet.universaldetector import UniversalDetector

__char_detector = UniversalDetector()


def detect_encoding(filename):
    __char_detector.reset()
    try:
        with open(filename, 'rb') as fp:
            for line in fp.readlines():
                __char_detector.feed(line)
                if __char_detector.done:
                    break
            result = __char_detector.result
            return result['encoding']
    except FileNotFoundError:
        return 'utf-8'


def normalize_path(path: Path, anchor=Path('.')) -> Path:
    path = Path(path)
    if path.is_absolute():
        return path.absolute()

    t = str(path)
    if t.startswith('~'):
        return Path(t.replace('~', str(Path.home()))).absolute()

    return anchor.absolute() / path


def quote_path(path, quote_char='"') -> str:
    path = str(path)
    return f'"{path}"' if ' ' in path else path



from .app_logger import AppLogger, LogLevel


class AbstractEnv:
    def __init__(self):
        self._logger = AppLogger()

    def print(self, *objs):
        self._logger.print(*objs, level=LogLevel.NORMAL)

    def debug(self, *objs):
        self._logger.print(*objs, level=LogLevel.DEBUG)

    def info(self, *objs):
        self._logger.print(*objs, level=LogLevel.INFO)

    def error(self, *objs):
        self._logger.print(*objs, level=LogLevel.ERROR)

    def warning(self, *objs):
        self._logger.print(*objs, level=LogLevel.WARNING)

    def critical(self, *objs):
        self._logger.print(*objs, level=LogLevel.CRITICAL)

    @property
    def log_level(self):
        return self._logger.level

    @log_level.setter
    def log_level(self,level):
        self._logger.level = level



from functools import lru_cache, total_ordering

from .timecode import *


@total_ordering
class TimePoint:
    def __init__(self, millisecond=0):
        self._millisecond = int(millisecond)

    @property
    def millisecond(self) -> int:
        return self._millisecond

    @property
    def second(self) -> float:
        return self._millisecond / 1000

    @property
    @lru_cache()
    def timestamp(self) -> TimeCode:
        parts = make_tc_parts(self._millisecond)
        return TimeCode(**parts)

    @classmethod
    def from_second(cls, second: float = 0):
        return cls(round(second * 1000))

    @classmethod
    def from_timestamp(cls, ts: str):
        assert is_tc(ts)
        tc_info = parse_tc_parts(ts)
        h, m, s, f = tc_info['tc_parts']
        hh = h * 60 * 60
        mm = m * 60
        second = hh + mm + s
        milli = second * 1000 + f
        return cls(milli)

    def __int__(self):
        return self._millisecond

    def __add__(self, other):
        return TimePoint(self._millisecond + int(other))

    def __iadd__(self, other):
        self._millisecond += int(other)

    def __isub__(self, other):
        self._millisecond -= int(other)

    def __sub__(self, other):
        return TimePoint(self._millisecond - int(other))

    def __eq__(self, other):
        return self._millisecond == int(other)

    def __lt__(self, other):
        return self._millisecond < int(other)

    def __hash__(self):
        return hash(self._millisecond)

    def __repr__(self):
        return f'TimePoint({self.millisecond})'



import pkgutil
from argparse import ArgumentParser

from rich.columns import Columns
from rich.markdown import Markdown
from rich.panel import Panel
from rich.text import Text
from rich_argparse import RichHelpFormatter

from cx_core import AbstractApp, DataPackage, LogLevel
from cx_core.folder_expander import FolderExpander
from cx_subtitle import SubtitleFormatter
from cx_subtitle.loader import *
from cx_subtitle.saver import *
from .env import env
from .exceptions import *


class SubConvApp(AbstractApp):
    __loader_classes = {
        '.srt': SrtLoader,
        '.txt': TxtLoader,
        '.docx': WordLoader,
        '.ttml': TTMLLoader
    }

    __saver_classes = {
        'srt': SrtSaver,
        'txt': TxtSaver,
        'word': WordSaver,
        'excel': ExcelSaver,
        'csv': CsvSaver
    }

    __suffixes = {
        'txt': '.txt',
        'word': '.docx',
        'excel': '.xlsx',
        'csv': '.csv',
        'srt': '.srt',
        'ttml': '.ttml',
        'rtf': '.rtf'
    }

    def __init__(self):
        super(SubConvApp, self).__init__()

        _parser = ArgumentParser(prog='sub_conv',
                                 description=f'批量提取台词本，支持的格式包括{" ".join(str(x) for x in self.__loader_classes.keys())}',
                                 epilog='Designed by Cxalio',
                                 formatter_class=RichHelpFormatter, exit_on_error=False)
        _parser.add_argument('--format', '-f', dest='format', default='txt',
                             choices=self.__saver_classes.keys(),
                             help='指定输出格式')

        _parser.add_argument('--encoding', '-c', dest='encoding', default='auto',
                             help='设定输出编码')

        _parser.add_argument('-o', '--output-folder', dest='target_folder', default=None,
                             help='制定输出目录')

        _parser.add_argument('-t', '--time', dest='keep_time', action='store_true',
                             help='尽量保留时间信息')

        _parser.add_argument('--bypass-formatter', dest='bypass_formatter', default=False,
                             action='store_true', help='跳过默认的内容检查器，使用原样输出')

        _parser.add_argument('--debug', '-d', action='store_true', dest='debug',
                             help='调试模式')

        _parser.add_argument('--dry-run', dest='dry_run', action='store_true',
                             help='干转模式，不执行写入操作')

        _parser.add_argument('--man', help='显示详细的说明',
                             dest='show_man', action='store_true')

        _parser.add_argument('sources', nargs='*',
                             help='需要转换的文件')

        self.parser = _parser
        self.args = None

    def __enter__(self):
        _args = self.parser.parse_args()
        self.args = DataPackage(**vars(_args))
        env.log_level = LogLevel.DEBUG if self.args.debug else LogLevel.WARNING
        env.info('Initialized ...')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        env.info('Bye ~')
        return False

    def is_loadable(self, source):
        suffix = Path(source).suffix.lower()
        return suffix in self.__loader_classes

    def make_loader(self, source) -> AbstractSubtitleLoader:
        source = Path(source)
        env.info(f'开始构建 loader <- {source}')
        if not source.is_file():
            raise ValueError('这玩意儿不是文件')
        loader_class = self.__loader_classes.get(source.suffix.lower())
        loader = loader_class(source) if loader_class else None
        return loader

    @staticmethod
    def show_man():
        data = pkgutil.get_data('sub_conv', 'help.md').decode('utf-8')
        env.console.print(Markdown(data))

    @property
    def target_suffix(self):
        suffix = self.__suffixes.get(str(self.args.format).lower())
        return suffix if suffix else '.txt'

    @property
    def target_encoding(self):
        return 'utf-8' if self.args.encoding == 'auto' else self.args.encoding

    @staticmethod
    def print_subtitles(subtitles):
        panels = (
            Panel(Text(x.content, style='blue'),
                  title=Text(str(x.start.timestamp), style='yellow'),
                  subtitle=Text(str(x.end.timestamp), style='green'),
                  title_align='left',
                  subtitle_align='right',
                  expand=True)
            for x in subtitles)
        env.print(Columns(panels, expand=True, equal=True))

    def make_saver(self, source) -> AbstractSubtitleSaver:
        source = Path(source)
        target = source.with_suffix(self.target_suffix)
        env.info(f'开始构建 saver -> {target.name}')
        if self.args.target_folder:
            target = Path(self.args.target_folder) / target.name
            env.info(f'检测到指定的目标文件夹,已更新目标："{target.absolute()}"')
        saver_class = self.__saver_classes.get(self.args.format, TxtSaver)
        saver = saver_class(target, keep_time=self.args.keep_time)
        saver.encoding = self.target_encoding
        return saver

    def do_job(self, source):
        env.info(f'开始执行任务{source}')
        source = Path(source)
        if not self.is_loadable(source):
            env.error(f'{source}的格式无法读取，将跳过任务')
            return

        if source.suffix == self.target_suffix:
            env.error(f'[yellow]{source.name}[/yellow] 源文件和目标文件格式相同，跳过处理。')
            return

        with self.make_loader(source) as loader:
            subtitles = [x for x in loader.subtitles()]

        if not self.args.bypass_formatter:
            formatter = SubtitleFormatter()
            subtitles = formatter(subtitles)

        if self.args.debug:
            self.print_subtitles(subtitles)

        if self.args.dry_run:
            env.print(f'[yellow]{source.name}[/yellow] 已成功读取，[red]干转模式跳过输出[/red] ...')
            return

        with self.make_saver(source) as saver:
            target = saver.target
            for sbt in subtitles:
                saver.write(sbt)

        env.print(f'[yellow]{target.name}[/yellow] ... [green]转换完成[/green]')

    def sources(self):
        if not self.args.sources:
            return []
        s = self.args.sources
        expander = FolderExpander()
        expander.suffixes_whitelist = self.__loader_classes.keys()
        yield from expander.expand(*s)

    def run(self):
        env.info(f'解析的参数：', self.args)

        if self.args.show_man:
            self.show_man()
            return

        if not self.args.sources:
            raise NoSourceError()

        if self.args.bypass_formatter:
            env.print('[red]旁通格式化器[/red] ...')

        with env.console.status('开始执行任务') as status:
            for s in self.sources():
                source = Path(s)
                status.update(f'{source.name} ...')
                self.do_job(source)


def run():
    with SubConvApp() as app:
        app.run()



import asyncio
import json
import signal
from pathlib import Path

from ffmpeg import Progress
from ffmpeg.asyncio import FFmpeg
from ffmpeg.errors import FFmpegError

from cx_core import DataPackage
from .env import env
from functools import cache


def make_target(source: Path, profile=None) -> str:
    if not profile:
        profile = env.profile
    result = profile.output.folder / source.name
    parent_level = profile.output.keep_parent_level
    if parent_level > 0:
        parents = source.absolute().parent.parts
        selected_parts = parents[-1 * parent_level:]
        t_folder = Path(*selected_parts)
        env.debug(
            f'取用 {parent_level} 个上级目录:\n{source.name} -> {t_folder}'
        )
        result = profile.output.folder / t_folder / source.name

    t_suffix = profile.output.suffix
    if not t_suffix.startswith('.'):
        t_suffix = '.' + t_suffix

    result = result.with_suffix(t_suffix)
    return str(result)


class Transcoder:
    def __init__(self, source: str, profile=None):
        self.task = None
        self.source = Path(source)
        self.profile = profile if profile else env.profile
        self.duration = None

    def __enter__(self):
        self.task = env.progress.add_task(description=f'{self.source.name}...')
        ffprobe = FFmpeg(executable='ffprobe').input(self.source, print_format='json', show_format=None)
        media = json.loads(asyncio.run(ffprobe.execute()))
        media = DataPackage(**media)
        self.duration = float(media.format.duration)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        env.progress.remove_task(self.task)
        self.task = None
        env.current_ffmpeg = None
        return False

    @property
    @cache
    def target(self):
        return make_target(self.source, self.profile)

    async def transcode(self):
        target = self.target
        parent = Path(target).absolute().parent
        parent.mkdir(parents=True, exist_ok=True)

        ffmpeg = FFmpeg()
        for k, v in self.profile.general.options.items():
            ffmpeg.option(k, v)

        if self.profile.general.hardware_accelerate:
            ffmpeg.option('hwaccel', self.profile.general.hardware_accelerate)

        if self.profile.general.overwrite_existed:
            ffmpeg.option('y')
        else:
            ffmpeg.option('n')

        ffmpeg.input(self.source, **self.profile.input.options)
        ffmpeg.output(target, **self.profile.output.options)

        @ffmpeg.on('progress')
        def on_progress(progress: Progress):
            desc = f'{self.source.name}... [yellow]x{round(progress.speed, 2)}[/yellow]'
            curr = progress.time.seconds
            env.progress.update(self.task, description=desc, completed=curr, total=self.duration)

        @ffmpeg.on('stderr')
        def on_stderr(line):
            env.debug(f'FFMPEG输出：{line}')

        @ffmpeg.on('start')
        def on_start(arguments):
            env.debug(f'开始执行任务: {arguments}')

        @ffmpeg.on('completed')
        def on_completed():
            env.debug(f'{self.source}转完了')

        @ffmpeg.on('terminated')
        def on_terminated():
            env.info('FFMPEG 终止')
            Path(target).unlink(True)
            env.print('尝试移除未完成的目标文件...')

        env.debug(ffmpeg.arguments)
        env.current_ffmpeg = ffmpeg
        await ffmpeg.execute()
        env.current_ffmpeg = None

    def run(self):
        try:
            asyncio.run(self.transcode())
        except FFmpegError:
            env.error('当前 FFmpeg 进程被终止')




import re
from enum import StrEnum


class Timebase:
    def __init__(self, rate, drop_frame=None):
        self.fps = int(round(rate))
        self.drop_frame = drop_frame if drop_frame is not None else self.fps == float(rate)
        self.input = rate

    def __repr__(self):
        return 'Timebase({0}{1})'.format(self.fps, 'DF' if self.drop_frame else '')

    def __int__(self):
        return self.fps


class TCMode(StrEnum):
    TC = "time_code"
    TS = "time_stamp"


class TimeCode:
    __TEMPLATES = {TCMode.TS: '{0[0]:0>2}:{0[1]:0>2}:{0[2]:0>2}{sep}{0[3]:0>3}',
                   TCMode.TC: '{0[0]:0>2}:{0[1]:0>2}:{0[2]:0>2}{sep}{0[3]:0>2}'}

    def __init__(self, tc_parts: list[int] = None, drop_frame=False, mode=TCMode.TS, custom_sep=None):
        self.__parts = tc_parts if tc_parts else [0, 0, 0, 0]
        self.drop_frame = drop_frame
        self.mode = mode
        self.custom_sep = custom_sep

    @property
    def sep(self):
        if self.custom_sep is not None:
            return self.custom_sep
        if self.mode is TCMode.TS:
            return ','
        return ';' if self.drop_frame else ':'

    @property
    def segments(self):
        return tuple(self.__parts)

    def __str__(self):
        return TimeCode.__TEMPLATES[self.mode].format(self.__parts, sep=self.sep)

    def __repr__(self):
        return 'TimeCode{0}{1}'.format(self.__parts, 'DF' if self.drop_frame else '')

    def __rich__(self):
        sep = '[yellow]:[/yellow]'
        sep2 = f'[yellow]{self.sep}[/yellow]'
        hh, mm, ss, ff = (f'[green]x[/green]' for x in self.__parts)
        return f'{hh}{sep}{mm}{sep}{ss}{sep2}{ff}'


TC_PATTERN = re.compile(r'(\d\d):(\d\d):(\d\d)([:;,\.])(\d\d\d?)')


def is_tc(string: str):
    return TC_PATTERN.fullmatch(string) is not None


_EMPTY_TC_INFO_ = {'tc_parts': (0, 0, 0, 0), 'drop_frame': False, 'mode': TCMode.TS}


def parse_tc_parts(tc: str):
    match = TC_PATTERN.fullmatch(tc.strip())
    if match is None:
        return _EMPTY_TC_INFO_
    groups = match.groups()
    sep = groups[3]
    parts = [int(groups[x]) for x in [0, 1, 2, 4]]
    return {
        'tc_parts': parts,
        'drop_frame': sep == ';',
        'mode': TCMode.TC if sep in ':;' else TCMode.TS
    }


def make_tc_parts(millisecond: int, mode=TCMode.TS, timebase: Timebase = None):
    assert millisecond >= 0
    seconds = millisecond // 1000
    ss = seconds % 60
    minutes = seconds // 60
    mm = minutes % 60
    hours = minutes // 60
    hh = hours % 24
    fff = millisecond % 1000
    if mode is TCMode.TC:
        if timebase is None:
            timebase = Timebase(24, False)
        rate = timebase.fps
        fff = int((millisecond / 1000) * rate) % rate
    tc_parts = [int(x) for x in [hh, mm, ss, fff]]
    return {'tc_parts': tc_parts, 'drop_frame': timebase.drop_frame if timebase else False,
            'mode': mode}



import csv
from abc import abstractmethod
from pathlib import Path

from docx import Document
from openpyxl import Workbook

from .subtitle import Subtitle


class AbstractSubtitleSaver:
    def __init__(self, filename, keep_time=False):
        self.target = Path(filename)
        self.keep_time = keep_time
        self.file = None
        self.encoding = 'utf-8'

    def __enter__(self):
        self.file = open(self.target, 'w', encoding=self.encoding)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.flush()
        self.file.close()
        return False

    @abstractmethod
    def write(self, subtitle: Subtitle):
        pass


class TxtSaver(AbstractSubtitleSaver):
    LINE_TEMPLATE = {
        True: '{start}\t{content}\n',
        False: '{content}\n'
    }

    def __init__(self, target, keep_time=False):
        super(TxtSaver, self).__init__(target, keep_time)

    def write(self, subtitle: Subtitle):
        line = TxtSaver.LINE_TEMPLATE[self.keep_time].format(start=subtitle.start.timestamp, content=subtitle.content)
        self.file.write(line)


class SrtSaver(AbstractSubtitleSaver):
    LINE_TEMPLATE = '{number}\n{start} --> {end}\n{content}\n\n'

    def __init__(self, target, keep_time=True):
        super(SrtSaver, self).__init__(target, keep_time)
        self.__count = 1

    def write(self, subtitle: Subtitle):
        line = SrtSaver.LINE_TEMPLATE.format(
            number=self.__count,
            start=subtitle.start.timestamp,
            end=subtitle.end.timestamp,
            content=subtitle.content
        )
        self.file.write(line)
        self.__count += 1


class WordSaver(AbstractSubtitleSaver):
    LINE_TEMPLATE = {
        True: '{start}\t{content}',
        False: '{content}'
    }

    def __init__(self, target, keep_time=False):
        super(WordSaver, self).__init__(target, keep_time)
        self.document = Document()

    def __enter__(self):
        self.document.add_heading(Path(self.target.stem).name, level=1)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.document.save(self.target)
        return False

    def write(self, subtitle: Subtitle):
        line = self.LINE_TEMPLATE[self.keep_time].format(
            start=subtitle.start.timestamp,
            content=subtitle.content
        )
        self.document.add_paragraph(line)


class ExcelSaver(AbstractSubtitleSaver):
    """如果 keep_time 为 True，则会追加秒数队列"""
    HEADERS = {
        True: ['开始', '结束', '开始（秒）', '结束（秒）', '台词'],
        False: ['开始', '结束', '台词']
    }

    def __init__(self, target, keep_time=True):
        super(ExcelSaver, self).__init__(target, keep_time)
        self.workbook = Workbook()
        self.table = self.workbook.active

    def __enter__(self):
        self.table.title = '字幕表格'
        self.table.append(self.HEADERS[self.keep_time])
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.workbook.save(self.target)
        return False

    def write(self, subtitle: Subtitle):
        row = [str(subtitle.start.timestamp), str(subtitle.end.timestamp)]
        if self.keep_time:
            row.append(subtitle.start.second)
            row.append(subtitle.end.second)
        row.append(subtitle.content)
        self.table.append(row)


class CsvSaver(AbstractSubtitleSaver):
    """如果 keep_time 为 True，则会追加秒数队列"""
    HEADERS = {
        True: ['开始', '结束', '开始（秒）', '结束（秒）', '台词'],
        False: ['开始', '结束', '台词']
    }

    def __init__(self, target, keep_time=False):
        super(CsvSaver, self).__init__(target, keep_time)
        self.__saver = None

    def __enter__(self):
        self.file = open(self.target, 'w', newline='', encoding=self.encoding)
        self.__saver = csv.DictWriter(self.file, self.HEADERS[self.keep_time], extrasaction='ignore')
        self.__saver.writeheader()
        return self

    def write(self, subtitle: Subtitle):
        start, end = subtitle.start, subtitle.end
        row = {
            '开始': str(start.timestamp),
            '结束': str(end.timestamp),
            '开始（秒）': start.second,
            '结束（秒）': end.second,
            '台词': subtitle.content.strip()
        }
        self.__saver.writerow(row)



import tomllib
from collections.abc import Iterable
from pathlib import Path

from cx_core import DataPackage
from cx_core.utils import normalize_path
from .env import env


class ProfileNoFoundError(FileNotFoundError):
    def __init__(self, target=None):
        super(ProfileNoFoundError, self).__init__()
        self._tgt = target

    def __str__(self):
        if self._tgt:
            return f'ffmpeg 配置文件 "{self._tgt}" 未找到，程序无法执行'
        return '未指定配置文件，程序无法运行'


class ProfileLoader:
    def __init__(self):
        self.task_id = None

    def __enter__(self):
        self.task_id = env.progress.add_task(description='初始化配置信息', start=False)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        env.progress.remove_task(self.task_id)
        self.task_id = None

    @staticmethod
    def __make_sure_list(a=None):
        if not a:
            return []
        if isinstance(a, str):
            return a.split(' ')
        if isinstance(a, Iterable):
            return [str(x) for x in a]
        return [x for x in str(a).split(' ')]

    @staticmethod
    def __check_options(a=None):
        if not a:
            return {}
        if isinstance(a, str):
            return ProfileLoader.__arrange_options(a.split(' '))
        if isinstance(a, dict | DataPackage):
            return {**a}
        if isinstance(a, list):
            return ProfileLoader.__arrange_options(a)
        return ProfileLoader.__arrange_options(str(a).split(' '))

    @staticmethod
    def __arrange_options(a: list) -> dict:
        result = {}
        previous_key = None
        for x in a:
            word = str(x)
            if word.startswith('-'):
                if previous_key:
                    result[previous_key] = None
                    previous_key = None
                else:
                    previous_key = word[1:]
            else:
                if previous_key:
                    result[previous_key] = word
                    previous_key = None
                else:
                    env.debug(f'已忽略无法识别的参数{word}')
        if previous_key:
            result[previous_key] = None
        return result

    def __check_data_package(self, package):
        env.progress.update(self.task_id, description='解析配置文件数据', completed=1, total=2)
        env.info('检查配置数据...')

        result = DataPackage(**package)
        env.debug(f'配置文件：[cyan]{result.general.name}[/cyan]:[yellow]{result.general.description}[/yellow]')

        if not result.general.working_folder:
            env.debug('配置文件未指定工作目录')
            profile_folder = normalize_path('.' if not result.profile_path else result.profile_path.absolute().parent)
            result.general.working_folder = profile_folder
        env.debug(f'general.working_folder 已设置为 {result.general.working_folder}')

        env.info('开始格式化特殊设置')
        result.general.options = ProfileLoader.__check_options(result.general.options)
        result.input.options = ProfileLoader.__check_options(result.input.options)
        result.output.options = ProfileLoader.__check_options(result.output.options)
        result.input.files = ProfileLoader.__make_sure_list(result.input.files)
        result.input.suffix_includes = ProfileLoader.__make_sure_list(result.input.suffix_includes)
        result.input.suffix_excludes = ProfileLoader.__make_sure_list(result.input.suffix_excludes)

        env.debug('增加手动指定的源文件')
        result.input.files.extend(ProfileLoader.__make_sure_list(env.args.sources))

        if not result.output.folder:
            env.debug('未设置目标文件夹')
            result.output.folder = Path('.')
        result.output.folder = normalize_path(result.output.folder, result.general.working_folder)

        return result

    def load(self, filename):
        result = {}
        env.progress.start_task(self.task_id)
        env.progress.update(self.task_id, completed=0, total=2)
        task = env.progress.add_task(description=f'读取{filename}...')
        try:
            with env.progress.open(filename, 'rb', task_id=task) as fp:
                result.update(tomllib.load(fp))
            result['profile_path'] = Path(filename)
            return self.__check_data_package(result)
        except FileNotFoundError:
            raise ProfileNoFoundError(filename)
        finally:
            env.progress.remove_task(task)



import itertools
from itertools import groupby
from pathlib import Path

from cx_core.folder_expander import FolderExpander
from .env import env
from .source_adapter import adapters


class Planner:
    _DEFAULT_SUFFIXES = ("mov mkv mp4 flv 3gp 3gpp rmvb mp3 aac"
                         " mxf mxf_op1a vob wmv wma srt ass aas ttml ogg oga ogv m4a"
                         " m4v 3g2 mpeg mpg ts lrc h264 flac ast asf gif")

    def __init__(self):
        self._loaded_file = set()
        self.plans = []
        env.debug('Planner 初始化')

    def __len__(self):
        return len(self.plans)

    def __iter__(self):
        return self.plans.__iter__()

    def load_source(self, filename):
        source = Path(filename)
        if not source.is_absolute():
            source = env.profile.general.working_folder / source
        suffix = source.suffix.lower().strip('.')
        env.debug(f'文件 "{source}" 的扩展名为 "{suffix}"')
        if suffix in adapters:
            env.debug(f'已识别 "{source}" 为列表文件')
            try:
                with adapters[suffix](source) as adp:
                    for item in adp.items():
                        env.debug(f'发现文件路径 "{item}"')
                        self.plans.append(item)
            except FileNotFoundError:
                env.error(f'文件 {source} 不存在，无法读取')
            finally:
                self._loaded_file.add(source)
        else:
            self.plans.append(source)
            self._loaded_file.add(source)
            env.debug(f'文件 "{source}" 无法识别为列表文件，将直接添加进入计划')

    def load_sources(self, filenames):
        task = env.progress.add_task(description='读取来源文件...')
        for f in env.progress.track(filenames, task_id=task):
            self.load_source(f)
        env.progress.remove_task(task)

    def arrange(self, profile=None):
        if not profile:
            profile = env.profile

        env.info('开始整理计划列表')

        env.debug('构建 Expander ...')
        expander = FolderExpander()
        expander.working_directory = profile.general.work_folder

        env.debug('构建扩展名白名单')
        _basic_suffixes = {str(x).strip('.') for x in
                           itertools.chain(Planner._DEFAULT_SUFFIXES.split(), profile.input.suffix_includes)}
        _blacklist = {str(x).strip('.') for x in profile.input.suffix_excludes}
        w_list = _basic_suffixes - _blacklist
        env.debug('生成扩展名白名单 [cyan]{" ".join(w_list)}[/cyan]')
        expander.suffixes_whitelist = ['.' + x for x in w_list]

        env.info('开始展开文件夹')
        new_files = [Path(x) for x in expander.expand(*self.plans)]
        env.info('正在根据路径进行排序')
        new_files = sorted(new_files, key=lambda a: a.absolute())

        env.info('开始逐个检查文件')
        arranged_files = []
        task = env.progress.add_task(description='检查文件列表...')
        for k, v in env.progress.track(groupby(new_files), task_id=task):
            kk = Path(k)
            if not kk.exists():
                env.warning(f'文件 "{kk}" 不存在，将从任务列表中去除')
                continue
            arranged_files.append(kk)
            env.debug(f'重新添加 "{kk}"')
            count = len(list(v))
            if count > 1:
                env.debug(f'发现 {count} 个重复的 "{k}" ，已经去除')
        env.progress.remove_task(task)
        self.plans = arranged_files
        env.info('计划列表自动整理完毕')



from dataclasses import dataclass
from functools import cached_property

from cx_core.timepoint import TimePoint


@dataclass(order=True)
class Subtitle:
    start: TimePoint
    end: TimePoint
    content: str

    def __rich__(self):
        return f'''[yellow]{self.start.timestamp} [cyan]-->[/cyan] {self.end.timestamp}[/yellow]\
[cyan] : [/cyan][blue]{self.content}[blue]'''

    @cached_property
    def duration(self):
        return self.end - self.start



from pathlib import Path
from .utils import normalize_path


class FolderExpander:
    def __init__(self, **kwargs):
        self.suffixes_blacklist = []
        self.suffixes_whitelist = []
        self.working_directory = '.'
        self.existed_only = True
        self.follow_symlinks = True
        self.accept_files = True
        self.accept_folders = True
        if kwargs is not None:
            self.__dict__.update(kwargs)

    def __make_absolute(self, path: Path):
        return normalize_path(path,self.working_directory)

    def __is_acceptable_file(self, path: Path):
        if not self.accept_files:
            return False
        if not path.is_file():
            return False
        if self.existed_only and not path.exists():
            return False
        suffix = path.suffix.lower()
        if self.suffixes_whitelist:
            return suffix in self.suffixes_whitelist
        return suffix not in self.suffixes_blacklist

    def __is_acceptable_folder(self, path: Path):
        if not self.accept_folders:
            return False
        return path.is_dir()

    def expand(self, *entries):
        for entry in entries:
            p = self.__make_absolute(entry)
            if self.__is_acceptable_file(p):
                yield p
            elif self.__is_acceptable_folder(p):
                for sub in p.iterdir():
                    yield from self.expand(p / sub)



from pathlib import Path
from .transcoder import make_target
from .env import env


def quote(filename) -> str:
    filename = str(filename)
    if filename.startswith('"') and filename.endswith('"'):
        return filename
    return '"' + filename + '"'


class ScriptWriter:
    def __init__(self, target, profile=None):
        self.target = Path(target).absolute()
        self.output = None
        self.planned_folders = set()
        self.ffmpeg = 'ffmpeg'
        self.profile = profile if profile else env.profile

    def __enter__(self):
        self.output = open(self.target, 'w')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.output.flush()
        self.output.close()
        return False

    def global_options(self):
        for k, v in self.profile.general.options.items():
            yield '-' + k
            if v:
                yield v
        if self.profile.general.hardware_accelerate:
            yield '-hwaccel'
            yield self.profile.general.hardware_accelerate
        if self.profile.general.overwrite_existed:
            yield '-y'
        else:
            yield '-n'

    def input_options(self):
        for k, v in self.profile.input.options.items():
            yield '-' + k
            if v:
                yield v

    def output_options(self):
        for k, v in self.profile.output.options.items():
            yield '-' + k
            if v:
                yield v

    def write(self, source: Path | str):
        line = [self.ffmpeg]
        target = make_target(Path(source), self.profile)
        folder = Path(target).parent

        if not folder.exists() and folder not in self.planned_folders:
            self.output.write(f'mkdir -p {quote(folder.absolute())}\n')
            env.debug(f'写入新建文件夹{folder}的任务')
            self.planned_folders.add(folder)

        for i in self.global_options():
            line.append(i)

        for i in self.input_options():
            line.append(i)

        line.append('-i')
        line.append(quote(source.absolute()))

        for i in self.output_options():
            line.append(i)

        line.append(quote(target))
        cmd = " ".join(line)
        self.output.write(f'{cmd}\n')
        env.debug(f'写入命令{cmd}')



import re
from abc import abstractmethod
from os import PathLike
from pathlib import Path

from docx import Document
from functools import cached_property

from cx_core.timepoint import TimePoint
from cx_core.utils import detect_encoding
from .subtitle import Subtitle


class AbstractSubtitleLoader:
    def __init__(self, source: str | PathLike):
        self.source = Path(source)
        self.file = None

    @cached_property
    def encoding(self):
        return detect_encoding(self.source)

    def __enter__(self):
        self.file = open(self.source, 'r', encoding=self.encoding)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.close()
        return False

    @abstractmethod
    def subtitles(self):
        pass


class TxtLoader(AbstractSubtitleLoader):
    MILLISECOND_PER_CHAR = 120
    MIN_DURATION_PER_LINE = 2000
    TS_PATTERN = r'\s*\d{2}:\d{2}:\d{2}.\d{3}\s*'

    def __init__(self, filename):
        super(TxtLoader, self).__init__(filename)
        self.__last_point = TimePoint(0)

    def subtitles(self):
        for a in self.file:
            content = re.sub(self.TS_PATTERN, '', a)
            duration = len(content) * TxtLoader.MILLISECOND_PER_CHAR
            if duration < TxtLoader.MIN_DURATION_PER_LINE:
                duration = TxtLoader.MIN_DURATION_PER_LINE
            start = self.__last_point
            end = start + TimePoint(duration)
            yield Subtitle(start=start, end=end, content=content)
            self.__last_point = end


class SrtLoader(AbstractSubtitleLoader):
    HEAD_LINE = re.compile(r'^\d+$')
    TIME_LINE = re.compile(r'^(\d\d:\d\d:\d\d[,.]\d\d\d) --> (\d\d:\d\d:\d\d[,.]\d\d\d)$')

    def __int__(self, filename):
        super(SrtLoader, self).__init__(filename)

    def subtitles(self):
        content, start, end = '', None, None
        state = None
        for a in self.file:
            line = a.strip()
            if not state:
                if SrtLoader.HEAD_LINE.fullmatch(line):
                    content = ''
                    start = None
                    end = None
                    state = 'started'
            elif state == 'started':
                match = SrtLoader.TIME_LINE.fullmatch(line)
                if match:
                    start = TimePoint.from_timestamp(match.group(1))
                    end = TimePoint.from_timestamp(match.group(2))
                    state = 'timed'
            elif state == 'timed':
                if line:
                    content += line
                else:
                    yield Subtitle(start=start, end=end, content=content)
                    state = None


class WordLoader(AbstractSubtitleLoader):
    MILLISECOND_PER_CHAR = 120
    MIN_DURATION_PER_LINE = 2000
    TS_PATTERN = r'\s*\d{2}:\d{2}:\d{2}.\d{3}\s*'

    def __init__(self, filename):
        super(WordLoader, self).__init__(filename)
        self.__last_point = TimePoint(0)
        self.document = None

    def __enter__(self):
        self.document = Document(str(self.source.absolute()))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False

    def subtitles(self):
        for a in self.document.paragraphs:
            content = re.sub(self.TS_PATTERN, '', a.text)
            duration = len(content) * TxtLoader.MILLISECOND_PER_CHAR
            if duration < TxtLoader.MIN_DURATION_PER_LINE:
                duration = TxtLoader.MIN_DURATION_PER_LINE
            start = self.__last_point
            end = start + TimePoint(duration)
            yield Subtitle(start=start, end=end, content=content)
            self.__last_point = end


class TTMLLoader(AbstractSubtitleLoader):
    PATTERN = r'<p begin=\"(\d\d:\d\d:\d\d.\d\d\d)\" end=\"(\d\d:\d\d:\d\d.\d\d\d)\">(.*?)</p>\s'
    REPLACEMENTS = {
        '<br/>': '\n',
        '&quot;': '"'
    }

    def __init__(self, filename):
        super(TTMLLoader, self).__init__(filename)

    def subtitles(self):
        total = self.file.read()
        pat = re.compile(self.PATTERN, re.DOTALL)
        for match in pat.finditer(total):
            start = TimePoint.from_timestamp(match.group(1))
            end = TimePoint.from_timestamp(match.group(2))
            content = match.group(3)
            for k, v in self.REPLACEMENTS.items():
                content = content.replace(k, v)
            yield Subtitle(
                start=start, end=end, content=content
            )



from enum import IntEnum

from rich.console import Console
from rich.text import Text


class LogLevel(IntEnum):
    DEBUG = 0
    INFO = 1
    WARNING = 2
    ERROR = 3
    CRITICAL = 4
    NORMAL = 999


class AppLogger:
    DEFAULT_HEADER_STYLES = {LogLevel.DEBUG: 'green', LogLevel.INFO: 'cyan', LogLevel.WARNING: 'purple',
                             LogLevel.ERROR: 'magenta', LogLevel.CRITICAL: 'red bold', LogLevel.NORMAL: 'blue'}

    def __init__(self, console=None, level=LogLevel.WARNING, header='::'):
        self._console = Console() if console is None else console
        self.level = level
        self.header = header
        self.header_styles = AppLogger.DEFAULT_HEADER_STYLES

    def styled_header(self, level=LogLevel.NORMAL):
        return Text(self.header, style=self.header_styles[level])

    def force_print(self, *objects, level=LogLevel.NORMAL):
        self._console.print(self.styled_header(level), *objects,
                            # style=AppLogger.DEFAULT_HEADER_STYLES[level],
                            markup=True, highlight=True)

    def print(self, *objects, level=LogLevel.NORMAL):
        if level < self.level:
            return
        else:
            self.force_print(*objects, level=level)



import re
from pathlib import Path


class FFmpegProgressParser:
    DURATION_RX = re.compile(r"Duration: (\d{2}):(\d{2}):(\d{2})\.(\d{2})")
    CURRENT_RX = re.compile(r"time=(\d{2}):(\d{2}):(\d{2})\.\d{2}")
    FILENAME_RX = re.compile(r"from '(.*)':")
    FPS_RX = re.compile(r"(\d{2}\.\d{2}|\d{2}) fps")

    @staticmethod
    def _seconds(hours, minutes, seconds, ms=0):
        millisecond = float(ms) / 100
        ss = (int(hours) * 60 + int(minutes)) * 60 + int(seconds)
        return float(ss) + millisecond

    def __init__(self) -> None:
        self.started = False
        self.duration = None
        self.fps = None
        self.filename = None
        self.current = None

    def __call__(self, line: str):
        self.started = True
        if self.duration is None:
            self.duration = self.get_duration(line)

        if self.filename is None:
            self.filename = self.get_filename(line)

        self.fps = self.get_fps(line)
        self.current = self.get_current(line)

    def get_duration(self, line: str):
        search = self.DURATION_RX.search(line)
        if search is not None:
            return self._seconds(*search.groups())
        return None

    def get_fps(self, line: str):
        search = self.FPS_RX.search(line)
        if search is not None:
            return round(float(search.group(1)))
        return None

    def get_filename(self, line: str):
        search = self.FILENAME_RX.search(line)
        if search is not None:
            p = Path(search.group(1))
            return p.name
        return None

    def get_current(self, line: str):
        search = self.CURRENT_RX.search(line)
        if search is not None:
            return self._seconds(*search.groups())
        return None



class DataPackage:
    def __init__(self, **kwargs):
        self._data = {}
        for k, v in kwargs.items():
            self._data[str(k)] = DataPackage.__check_value(v)

    def __repr__(self):
        return repr(self._data)

    @staticmethod
    def __check_value(value):
        if isinstance(value, dict):
            return DataPackage(**value)
        if isinstance(value, list):
            return [DataPackage.__check_value(x) for x in value]
        return value

    def _get_value(self, child_key, *keys):
        child = self._data.get(child_key)
        if not keys:
            return child
        if isinstance(child, DataPackage):
            return child._get_value(*keys)
        return NotImplemented

    def _set_value(self, child_key, *keys, value=None):
        if not keys and child_key in self._data:
            self._data[child_key] = value
            return
        child = self._data[child_key]
        if isinstance(child, DataPackage):
            child._set_value(*keys, value=value)
            return
        raise KeyError('Invalid key path:', child_key, keys)

    def __getitem__(self, item):
        keys = str(item).split('.')
        return self._get_value(*keys)

    def __setitem__(self, key, value):
        keys = str(key).split('.')
        return self._set_value(*keys, value=DataPackage.__check_value(value))

    def __getattr__(self, item):
        keys = str(item).split('.')
        return self._get_value(*keys)

    def __setattr__(self, key, value):
        if str(key).startswith('_'):
            self.__dict__[key] = value
        else:
            self._data[str(key)] = DataPackage.__check_value(value)

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return self._data.__iter__()

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def items(self):
        return self._data.items()



class JobCounter:
    def __init__(self, total=None):
        self.total = total
        self.current = 0
        self.number_length = 4 if self.total is None else len(str(self.total))
        self.template = '({0:>?}/{1})'.replace('?', str(self.number_length))

    def __str__(self):
        return self.template.format(self.current, self.total)

    def reset(self):
        self.current = 0

    def advance(self,n=1):
        self.current += n




import pkgutil
from argparse import ArgumentParser, ArgumentError
from pathlib import Path

from rich.columns import Columns
from rich.markdown import Markdown
from rich.panel import Panel
from rich_argparse import RichHelpFormatter

from cx_core import AbstractApp
from cx_core import DataPackage
from cx_core.app_logger import LogLevel
from cx_core.utils import normalize_path
from .env import env
from .planner import Planner
from .profile_loader import ProfileLoader, ProfileNoFoundError
from .script_writer import ScriptWriter
from .transcoder import Transcoder

from cx_core.tui_elements import JobCounter


class MediaKillerApp(AbstractApp):
    def __init__(self):
        super(MediaKillerApp,self).__init__()
        self.global_task = None
        self.app_name = 'media-killer'
        self.app_version = '0.1.2'
        parser = ArgumentParser(prog=self.app_name, formatter_class=RichHelpFormatter,
                                description='简单批量转码工具', epilog='Designed by xiii_1991')
        parser.add_argument('profile', help='指定配置文件路径', default=None, nargs='?')
        parser.add_argument('-g', '--generate-example-profile',
                            action="store_true", dest='generate_example',
                            help='生成范例文件')
        parser.add_argument('-a', '--add-source',
                            action='append', dest='sources', metavar='SOURCE_FILE',
                            help='增加来源文件')
        parser.add_argument('-s', '--make-script',
                            dest='script_output', metavar='SCRIPT_OUTPUT',
                            help='生成对应的脚本文件')
        parser.add_argument('-d', '--debug',
                            action='store_true', dest='debug', help='显示调试信息')
        parser.add_argument('--dry-run',
                            dest='dry_run', action='store_true', help='空转模式，不执行命令')
        parser.add_argument('--man', help='显示详细的说明',
                            dest='show_man', action='store_true')
        parser.add_argument('-v', '--version', action='version', version=self.app_version,
                            help='显示软件版本信息')
        self._parser = parser
        self.args = None

    @staticmethod
    def show_man():
        data = pkgutil.get_data('media_killer', 'help.md').decode('utf_8')
        env.console.print(Markdown(data))

    @staticmethod
    def copy_example_profile(tgt):
        tgt = normalize_path(tgt)
        data = pkgutil.get_data('media_killer', 'example_project.toml')
        try:
            with open(tgt, 'xb') as file:
                file.write(data)
            env.print(f'模板配置文件已保存到 {env.args.profile} ，[red]请勿直接加载运行！[/red]')
        except FileExistsError:
            env.error('文件 {0} 已存在，禁止覆盖文件'.format(tgt))

    def __enter__(self):
        env.start()

        _args = self._parser.parse_args()
        self.args = DataPackage(**vars(_args))

        env.debug('解析命令行参数：', self.args)
        env.log_level = LogLevel.DEBUG if self.args.debug else LogLevel.WARNING
        env.args = self.args

        self.global_task = env.progress.add_task(description='全局进度', start=False, visible=False)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        env.progress.stop_task(self.global_task)
        env.progress.remove_task(self.global_task)
        self.global_task = None

        result = False
        if exc_type is None:
            pass
        elif issubclass(exc_type, ProfileNoFoundError):
            env.critical(exc_val)
            result = True
        elif issubclass(exc_type, ArgumentError):
            env.error('参数输入有误:', exc_val)
            result = True

        env.stop()
        return result

    def run(self):
        if env.args.show_man:
            MediaKillerApp.show_man()
            return

        if not env.args.profile:
            raise ProfileNoFoundError

        if env.args.generate_example:
            env.debug('检测到 "--generate-example" 参数，拷贝配置文件模板')
            env.progress.update(task_id=self.global_task, description='正在输出配置文件', visible=True)
            self.copy_example_profile(env.args.profile)
            return

        with ProfileLoader() as profile_loader:
            datas = profile_loader.load(env.args.profile)
            env.profile = datas
        env.info('配置信息已初始化')

        if not env.profile.input.files:
            env.error('未指定任何来源信息，无事可做')
            return

        planner = Planner()
        env.progress.update(task_id=self.global_task, description='正在将来源信息加入计划', visible=True)
        planner.load_sources(env.profile.input.files)
        env.info('已发现 {0} 项来源'.format(len(planner)))

        env.progress.update(task_id=self.global_task, description='正在整理来源信息', visible=True)
        planner.arrange(env.profile)
        env.print(f'整理结束，已添加 {len(planner)} 项计划')
        if env.args.debug:
            col = Columns(
                (Panel(Path(x).name, style='yellow', border_style='green', safe_box=True,
                       title=f'#{i}', title_align='left')
                 for i, x in enumerate(planner)),
                expand=True, equal=False, column_first=True
            )
            env.debug(col)

        if env.args.script_output is not None:
            env.progress.update(task_id=self.global_task, description='输出脚本文件', visible=True)
            env.debug('检测到 "--script" 参数，进行脚本输出')
            output_file = normalize_path(env.args.script_output)
            with ScriptWriter(output_file, env.profile) as writer:
                for plan in env.progress.track(planner, description=env.args.script_output):
                    writer.write(plan)
            env.print('已输出脚本文件 “{0}”'.format(env.args.script_output))
            return

        env.progress.update(self.global_task, description='总体进度', visible=True, total=len(planner))
        env.progress.start_task(self.global_task)

        job_counter = JobCounter(len(planner))
        for plan in planner:
            job_counter.advance()
            if env.wanna_quit:
                env.print('用户申请终止运行，取消未完成的任务')
                break
            if env.args.dry_run:
                env.print(plan)
            else:
                with Transcoder(plan, env.profile) as coder:
                    coder.run()
            env.progress.advance(self.global_task)
            env.print(f'[yellow]{job_counter}[/yellow] "{Path(plan).name}" [blue]...[/blue] FINISHED')
        env.print('任务执行结束')


def run():
    with MediaKillerApp() as media_killer:
        media_killer.run()



from rich.console import Console

from cx_core import AbstractEnv, AppLogger


class Env(AbstractEnv):
    def __init__(self):
        super(Env, self).__init__()
        self.console = Console()
        self._logger = AppLogger(console=self.console)


env = Env()



import re

from .subtitle import Subtitle


class SubtitleFormatter:
    def __init__(self, **kwargs):
        self.replacements = {'\t': ' '}
        self.remove_tags = True

        self.normal_strip = True
        self.strip_quotes = True
        self.extra_strips = ''

        self.remove_empty_lines = True
        self.shrink_long_spaces = True
        self.shrink_long_quotes = True

        self.remove_empty_subtitles = True
        self.remove_zero_duration_subtitles = True

        self.__dict__.update(kwargs)

    def __basic_content_cleaning(self, content: str):
        result = content
        for k, v in self.replacements.items():
            result = result.replace(k, v)
        result = re.sub(r'\r\n', r'\n', result)
        return result

    @staticmethod
    def __remove_tags(content: str):
        return re.sub(
            r'<(\w+)>([^<>]*)</\1>',
            r'\2',
            content
        )

    def __strip(self, content: str):
        result = str(content)
        while True:
            current = result
            if self.normal_strip:
                current = current.strip()
            if self.strip_quotes:
                current = current.strip('\'"‘“”’')
            if self.extra_strips:
                current = current.strip(self.extra_strips)
            if current == result:
                break
            result = current
            current = None
        return result

    @staticmethod
    def __remove_empty_lines(content: str):
        result = re.sub(r'\n\s+', r'\n', content)
        return result

    @staticmethod
    def __shrink_long_spaces(content: str):
        result = re.sub(r' +', r' ', content)
        return result

    @staticmethod
    def __shrink_long_quotes(content: str):
        result = re.sub(r'([\'\"‘“”’])+', r'\1', content)
        return result

    def format_subtitle(self, subtitle: Subtitle):
        content = self.__basic_content_cleaning(subtitle.content)
        content = self.__strip(content)
        if self.remove_tags:
            content = self.__remove_tags(content)
        if self.remove_empty_lines:
            content = self.__remove_empty_lines(content)
        if self.shrink_long_spaces:
            content = self.__shrink_long_spaces(content)
        if self.shrink_long_quotes:
            content = self.__shrink_long_quotes(content)
        subtitle.content = content
        return subtitle

    def is_subtitle_legal(self, subtitle: Subtitle):
        if self.remove_empty_subtitles and not subtitle.content:
            return False
        if self.remove_zero_duration_subtitles and subtitle.duration == 0:
            return False
        return True

    def format_sequence(self, subtitles: list[Subtitle]):
        result = [self.format_subtitle(sub) for sub in subtitles if self.is_subtitle_legal(sub)]
        result = sorted(result, key=lambda x: x.start)
        return result

    def __call__(self, sth: Subtitle | list[Subtitle]):
        if isinstance(sth, list):
            return self.format_sequence(sth)
        elif isinstance(sth, Subtitle):
            return self.format_subtitle(sth)
        return sth



import urllib.parse
from pathlib import Path
from cx_core.utils import detect_encoding
from abc import abstractmethod
import csv
from .env import env
import xml.etree.ElementTree as ET


class AbstractAdapter:
    def __init__(self):
        self.filename = None
        self.file = None
        self.task = None

    def codec(self):
        return detect_encoding(self.filename)

    def __enter__(self):
        self.file = open(self.filename, 'r', encoding=self.codec())
        self.task = env.progress.add_task(description=f'读取 {self.filename} ...')
        env.progress.wrap_file(self.file, task_id=self.task)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.close()
        env.progress.remove_task(self.task)
        return False

    @abstractmethod
    def items(self):
        pass


class CSVAdapter(AbstractAdapter):
    suffix = 'csv'

    def __init__(self, filename):
        super(AbstractAdapter, self).__init__()
        self.filename = Path(filename)

    def items(self):
        reader = csv.DictReader(self.file, fieldnames=['File Name', 'Clip Directory'])
        for row in reader:
            if row['Clip Directory'] == 'Clip Directory':
                env.debug('跳过标题行')
                continue
            source_file = Path(row['Clip Directory'], row['File Name'])
            yield source_file


class TxtAdapter(AbstractAdapter):
    suffix = 'txt'

    def __init__(self, filename):
        super(AbstractAdapter, self).__init__()
        self.filename = Path(filename)

    def items(self):
        for line in self.file:
            yield Path(line.strip().strip('"'))


class XmlAdapter(AbstractAdapter):
    suffix = 'xml'

    def __init__(self, filename):
        super(AbstractAdapter, self).__init__()
        self.filename = filename

    def items(self):
        et = ET.parse(self.file)
        for e in et.iter('pathurl'):
            url = urllib.parse.unquote_plus(e.text)
            if not url.startswith('file:'):
                continue
            path = url.removeprefix('file:')
            yield Path(path)


class FcpXmlAdapter(AbstractAdapter):
    suffix = 'fcpxml'

    def __init__(self, filename):
        super(AbstractAdapter, self).__init__()
        self.filename = filename

    def items(self):
        et = ET.parse(self.file)
        for e in et.iter('media-rep'):
            url = urllib.parse.unquote_plus(e.attrib['src'])
            if not url.startswith('file:'):
                continue
            path = url.removeprefix('file:')
            yield Path(path)


adapters = {'csv': CSVAdapter, 'txt': TxtAdapter, 'xml': XmlAdapter, 'fcpxml': FcpXmlAdapter}



import signal
import asyncio

from rich.progress import Progress, SpinnerColumn, MofNCompleteColumn

from cx_core import AppLogger, AbstractEnv, LogLevel, DataPackage


class Env(AbstractEnv):
    def __init__(self):
        super(Env,self).__init__()
        self._progress = Progress(
            SpinnerColumn(),
            # MofNCompleteColumn(),
            *Progress.get_default_columns(),
            expand=True,
            transient=True)

        self._logger = AppLogger(console=self._progress.console)
        self._logger.level = LogLevel.WARNING

        self.args = None
        self.profile = DataPackage()

        self.current_ffmpeg = None
        self.wanna_quit = False

    def start(self):
        self._progress.start()
        self.info('Started ...')

    def stop(self):
        self._progress.stop()
        self.info('Bye~')

    @property
    def progress(self):
        return self._progress

    @property
    def console(self):
        return self._progress.console


env = Env()


def signal_handler(sig, frame):
    if sig != signal.SIGINT:
        return
    env.info('接收到 SIGINT')
    env.print("收到终止信号，准备退出...")
    env.wanna_quit = True
    if env.current_ffmpeg:
        env.current_ffmpeg.terminate()


signal.signal(signal.SIGINT, signal_handler)
